import struct
# noinspection PyProtectedMember
from enum import Enum, Flag, _decompose  # yes this method is meant to be package private but Flag did everything I wanted it to apart from this
from datetime import timedelta, datetime, timezone


class Mode(Enum):
    """Translates api return to osu! gamemode"""
    STANDARD = 0
    TAIKO = 1
    CTB = 2
    MANIA = 3


class Teams(Enum):
    """Translates api return to osu! multiplayer team"""
    NONE = 0
    BLUE = 1
    RED = 2


class TeamType(Enum):
    """Translates api return to osu! multiplayer team type"""
    HEADTOHEAD = 0
    TAGCOOP = 1
    TEAMVS = 2
    TAGTEAMVS = 3


class WinCon(Enum):
    """Translates api return to osu! multiplayer scoring type"""
    SCORE = 0
    ACCURACY = 1
    COMBO = 2
    SCOREV2 = 3


class Approval(Enum):
    """Translates api return to a stage in the map ranking process"""
    GRAVEYARD = -2
    WIP = -1
    PENDING = 0
    RANKED = 1
    APPROVED = 2
    QUALIFIED = 3
    LOVED = 4


class Genre(Enum):
    """Translates an api return to a specific genre"""
    ANY = 0
    UNSPECIFIED = 1
    VIDEO_GAME = 2
    ANIME = 3
    ROCK = 4
    POP = 5
    OTHER = 6
    NOVELTY = 7
    HIP_HOP = 9
    ELECTRONIC = 10
    METAL = 11
    CLASSICAL = 12
    FOLK = 13
    JAZZ = 14


class Language(Enum):
    """Translates an api return to a specific language"""
    ANY = 0
    UNSPECIFIED = 1
    ENGLISH = 2
    JAPANESE = 3
    CHINESE = 4
    INSTRUMENTAL = 5
    KOREAN = 6
    FRENCH = 7
    GERMAN = 8
    SWEDISH = 9
    SPANISH = 10
    ITALIAN = 11
    RUSSIAN = 12
    POLISH = 13
    OTHER = 14


mod_order = ["EZ", "HD", "DT", "NC", "HT", "HR", "FL", "SD", "PF", "NF", "SO", "TF", "NM"]
"""[str]: Ordered list of mods. Can probably override to customise print order."""


# sorry alex
# noinspection PyProtectedMember
class Mods(Flag):
    """Translate between bitwise api return and mod combinations for a score """
    def __str__(self):
        """Represents the mod when it is printed

        Returns:
            str: string representing the mods used in the "correct" (up to interpretation but I used scoreposts as a base) order
        """
        m = self.mod_list()
        return ''.join([mod for mod in mod_order if mod in m])

    def mod_list(self):
        """Get a list of the mods used as strings

        Returns:
            [str]: mods used as strings names
        """
        members, uncovered = _decompose(self.__class__, self._value_)
        mods = [m._name_ for m in members]
        if "DT" in mods and "NC" in mods:
            mods.remove("DT")
        if "SD" in mods and "PF" in mods:
            mods.remove("SD")
        return mods

    NM = 0
    NF = 1
    EZ = 2
    TD = 4
    HD = 8
    HR = 16
    SD = 32
    DT = 64
    RX = 128
    HT = 256
    NC = 512  # Only set along with DoubleTime. i.e: NC only gives 576
    FL = 1024
    Autoplay = 2048
    SO = 4096
    Relax2 = 8192  # Autopilot
    PF = 16384  # Only set along with SuddenDeath. i.e: PF only gives 16416
    KEY4 = 32768
    KEY5 = 65536
    KEY6 = 131072
    KEY7 = 262144
    KEY8 = 524288
    FadeIn = 1048576
    Random = 2097152
    Cinema = 4194304
    Target = 8388608
    KEY9 = 16777216
    KEYCOOP = 33554432
    KEY1 = 67108864
    KEY3 = 134217728
    KEY2 = 268435456
    ScoreV2 = 536870912
    Mirror = 1073741824


def modstr_to_enums(mods: str):
    """Translate a string of mods into their respective enums

    Args:
        mods: string to parse

    Returns:
        [Mods]: list of mods used as enums
    """
    parsed_mods = []
    mods = mods.replace(" ", "")
    while mods:
        mod_str = ""
        for letter in mods:
            mod_str += letter
            try:
                parsed_mods.append(Mods[mod_str])
                break
            except KeyError:
                pass
        mods = mods[len(mod_str):]
    return parsed_mods


def parse_mod_string(mods: str):
    """Parse a mod string into one enum representing it.

    Args:
        mods: string of mods used (e.g. "HDDTHR")

    Returns:
        Mods: Mod enum representing the mod combination
    """
    return Mods(sum([x.value for x in modstr_to_enums(mods)]))


class Playtime:
    """Object holding some information about playtime. Can be printed to just show value.

    Attributes:
        seconds: seconds played
        hours: hours played
        days: days played
        combined: formatted time spent playing
    """
    def __init__(self, seconds):
        self.seconds = seconds
        self.hours = seconds / 60 / 60
        self.days = seconds / 60 / 60 / 24
        self.combined = timedelta(seconds=seconds)

    def __repr__(self):
        return f"{self.combined}"


def read_byte(db):
    return int.from_bytes(db.read(1), "little")


def read_short(db):
    return int.from_bytes(db.read(2), "little")


def read_int(db):
    data = db.read(4)
    return int.from_bytes(data, "little")


def read_long(db):
    return int.from_bytes(db.read(8), "little")


def read_single(db):
    return struct.unpack("<f", db.read(4))[0]


def read_double(db):
    return struct.unpack("<d", db.read(8))[0]


def read_bool(db):
    return struct.unpack("?", db.read(1))[0]


def read_leb128(db):
    result = 0
    shift = 0
    while True:
        byte = read_byte(db)
        result |= (byte & 0x7f) << shift
        if (byte & 0x80) == 0:
            break
        shift += 7
    return result


def read_string(db):
    present = db.read(1)
    if present == b'\x00':
        return None

    size = read_leb128(db)
    string = db.read(size)
    try:
        ret = string.decode('utf-8')
    except ValueError:
        ret = string.decode('ISO-8859-1')
    return ret


def read_int_double(db):
    fst = read_byte(db)
    item_1 = read_int(db)
    snd = read_byte(db)
    item_2 = read_double(db)
    return item_1, item_2


def read_timing(db):
    bpm = read_double(db)
    offset = read_double(db)
    inherited = read_bool(db)
    return bpm, offset, inherited


def read_datetime(db):
    ticks = read_long(db)
    start = datetime(year=2001, month=1, day=1, tzinfo=timezone.utc)
    return start + timedelta(microseconds=ticks*0.1)



from .utils import *


class BaseScore:
    """Base class for a score, all types of score share these attributes.

    Attributes:
        client: OsuClient that created this object instance
        map_id: the map the score is for
        num_300: the number of 300s hit in the score
        num_100: the number of 100s hit in the score
        num_50: the number of 50s hit in the score
        misses: the number of misses in the score
        max_combo: the highest combo reached by the player in this score
        num_katu: the number of katu hit
        num_geki: the number of geki hit
        perfect: boolean representing if the maximum combo for the map was reached
        user_id: id of the player who set the score
        score: the actual numerical score for the play
        username: name of the player who set the score
        successful_hits: number of non-misses
        map_total_hits: maximum number of non-misses
        accuracy_dec: the player's accuracy as a decimal
    """

    def __init__(self, score_info, client):
        self.client = client
        self.num_300 = int(score_info['count300'])
        self.num_100 = int(score_info['count100'])
        self.num_50 = int(score_info['count50'])
        self.misses = int(score_info['countmiss'])
        self.max_combo = int(score_info['maxcombo'])
        self.num_katu = int(score_info['countkatu'])
        self.num_geki = int(score_info['countgeki'])
        self.perfect = (score_info['perfect'] == "1") or score_info['perfect']
        self.user_id = int(score_info['user_id']) if "user_id" in score_info else None
        self.score = int(score_info['score'])
        self.username = score_info['username'] if "username" in score_info else self.user_id
        self.successful_hits = self.num_50 + self.num_100 + self.num_300
        self.map_total_hits = self.successful_hits + self.misses
        temp_accuracy = (self.num_50 * 50 + self.num_100 * 100 + self.num_300 * 300) / (
                self.map_total_hits * 300) if self.map_total_hits != 0 else 0
        self.accuracy_dec = min(1.0, max(0.0, temp_accuracy))

    def fetch_user(self):
        """Make an api call to get more information about user.

        Returns:
            User: the user who set the score
        """
        return self.client.fetch_user(user_id=self.user_id)


class Score(BaseScore):
    """Generic score object, returned when searching or getting best plays.

    Attributes:
        score_id: the id of the score
        timestamp: when the score was set
        mods: enum representing the mod combination used
        rank: letter representing how good the score is
        pp: the amount of pp rewarded for the score
        replay_available: boolean representing if it is possible to download the replay of the map

    """

    def __init__(self, score_info, client, map_id):
        self.map_id = map_id
        self.score_id = int(score_info['score_id'])
        self.timestamp = datetime.strptime(score_info['date'], "%Y-%m-%d %H:%M:%S")
        self.mods = Mods(int(score_info['enabled_mods']))
        self.rank = score_info['rank']
        self.pp = float(score_info['pp'])
        self.replay_available = score_info['replay_available'] == "1"
        super().__init__(score_info, client)

    def __repr__(self):
        return f"{self.mods} score on beatmap {self.map_id} by {self.username}"


class RecentScore(BaseScore):
    """Score object used when getting recent scores.

    Attributes:
        timestamp: when the score was set
        mods: the mod combination used
        rank: letter representing how good the score is
    """

    def __init__(self, score_info, client, map_id):
        self.map_id = map_id
        self.timestamp = datetime.strptime(score_info['date'], "%Y-%m-%d %H:%M:%S")
        self.mods = Mods(int(score_info['enabled_mods']))
        self.rank = score_info['rank']
        super().__init__(score_info, client)

    def __repr__(self):
        return f"{self.mods} score on beatmap {self.map_id} by {self.username}"


class MultiScore(BaseScore):
    """Multiplayer lobby score object.

    Attributes:
        game_id: the game it was set in
        match_id: the match it ws set in
        slot: where in the lobby the player was positioned
        team: what team the player was on
        rank: letter achieved
        passed: boolean representing if the player passed or not
        mods: all the mods this player used, not necessarily the same as other scores from the same game

    """

    def __init__(self, score_info: dict, client, map_id: int, match_id: int, game_id: int, mods: Mods):
        self.map_id = map_id
        self.game_id = game_id
        self.match_id = match_id
        self.slot = int(score_info["slot"])
        self.team = Teams(int(score_info["team"]))
        self.rank = int(
            score_info["rank"])  # says not used on the wiki but lists it anyway so ill catch it just in case
        self.passed = score_info["pass"] == "1"
        self.mods = mods if not score_info["enabled_mods"] else Mods(mods.value + int(score_info["enabled_mods"]))
        super().__init__(score_info, client)

    def __repr__(self):
        return f"{self.mods} score on beatmap {self.map_id} by {self.user_id} in match {self.match_id}"


class LocalScore(BaseScore):
    def __init__(self, score_info, client, replay_hash):
        super().__init__(score_info, client)
        self.md5_hash = replay_hash
        self.mods = Mods(score_info["mods"])
        self.timestamp = score_info["timestamp"]  # todo make datetime
        self.online_id = score_info["online_id"]

    def __repr__(self):
        return f"{self.mods} score by {self.username}"



from datetime import datetime
from .utils import *
from .score import MultiScore


class Match:
    """A past/ongoing multiplayer lobby

    Attributes:
        client: OsuClient that created the instance
        match_id: id of the lobby
        name: the title of the lobby
        start_time: timestamp of when the match started
        end_time: timestamp of when the match ended (or None if it is ongoing)
        games: [Game] list of game objects containing info on each specific round of the lobby
        url: the url for the match to view in a browser
    """
    def __init__(self, match_info, client):
        match_attr = match_info['match']
        games_list = match_info['games']
        self.client = client
        self.match_id = int(match_attr['match_id'])
        self.name = match_attr['name']
        self.start_time = datetime.strptime(match_attr['start_time'], "%Y-%m-%d %H:%M:%S")
        self.end_time = datetime.strptime(match_attr['end_time'], "%Y-%m-%d %H:%M:%S") if match_attr["end_time"] else None
        self.games = [Game(game_info, client, self.match_id) for game_info in games_list]
        self.url = f"https://osu.ppy.sh/community/matches/{self.match_id}"

    def __repr__(self):
        games = "\n".join([str(x) for x in self.games])
        return f"{self.match_id}: {self.name}\n{games}"


class Game:
    """A specific round of a multiplayer lobby

    Attributes:
        game_id: the id of the round
        match_id: the id of the lobby it happened in
        client: OsuClient that created the instance
        start_time: timestamp of when the game started
        end_time: timestamp of when the game ended
        play_mode: the gamemode the game happened in
        map_id: the id of the map that was played
        match_type: doesn't seem to actually do anything? but the api returns it so I'm capturing it
        score_type: enum representing how the game was scored
        team_type: enum representing how the teams were set
        mods: enum representing the mods that were used
        scores: a list of multiplayer-specific score objects representing the results of the round for each player
    """
    def __init__(self, game_info, client, match_id: int):
        self.game_id = int(game_info['game_id'])
        self.match_id = match_id
        self.client = client
        self.start_time = datetime.strptime(game_info['start_time'], "%Y-%m-%d %H:%M:%S")
        self.end_time = datetime.strptime(game_info['end_time'], "%Y-%m-%d %H:%M:%S")
        self.play_mode = Mode(int(game_info['play_mode']))
        self.map_id = int(game_info['beatmap_id'])
        self.match_type = int(game_info['match_type'])
        self.score_type = WinCon(int(game_info['scoring_type']))
        self.team_type = TeamType(int(game_info['team_type']))
        self.mods = Mods(int(game_info['mods']))
        self.scores = [MultiScore(score_info, client, self.map_id, match_id, self.game_id, self.mods)
                       for score_info in game_info['scores']]

    def get_players(self):
        """The users who took part in this game. Not an api call.

        Returns:
            [int]: list of user id's
        """
        return [score.user_id for score in self.scores]

    def __repr__(self):
        return f"{self.game_id}: {self.score_type.name} {self.team_type.name} game with {self.mods} on map {self.map_id}"



from .osuclient import OsuClient
from .user import User
from .map import Map, LocalMap
from .score import Score, RecentScore, MultiScore, LocalScore
from .match import Match, Game
from .db import OsuDB, ScoresDB, Collections
from . import utils



import urllib.request
import pyttanko

from .utils import *


class BaseMap:
    def __init__(self, map_info, client):
        self.client = client
        self.song_title = map_info['title']
        self.artist = map_info['artist']
        self.length = float(map_info['total_length'])
        self.difficulty_name = map_info['version']
        self.mode = Mode(int(map_info['mode']))
        self.beatmap_id = int(map_info['beatmap_id'])
        self.mapset_id = int(map_info['beatmapset_id'])
        self.creator_name = map_info['creator']
        self.tags = map_info['tags'].split(" ")
        self.approval = Approval(int(map_info['approved']))
        self.circle_size = float(map_info['diff_size'])
        self.overall_difficulty = float(map_info['diff_overall'])
        self.approach_rate = float(map_info['diff_approach'])
        self.hp_drain = float(map_info['diff_drain'])
        self.source = map_info['source']
        self.circle_count = int(map_info['count_normal'])
        self.slider_count = int(map_info['count_slider'])
        self.spinner_count = int(map_info['count_spinner'])
        self.total_objects = self.circle_count + self.slider_count + self.spinner_count
        self.md5_hash = map_info['file_md5']

    def __repr__(self):
        return f"{self.song_title} [{self.difficulty_name}] mapped by {self.creator_name}"


class LocalMap(BaseMap):
    def __init__(self, map_info, client):
        self.filesize = map_info["file_size"]
        self.artist_unicode = map_info["artist_unicode"]
        self.title_unicode = map_info["title_unicode"]
        self.osu_version = map_info["osu_version"]
        self.audio_filename = map_info["audio_filename"]
        self.filename = map_info["filename"]
        self.last_mod_time = map_info["last_mod_time"]
        # self.last_mod_time_2 = map_info["last_mod_time_2"]
        self.slider_velocity = map_info["slider_velocity"]
        if int(self.osu_version) >= 20140609:
            self.standard_sr_ratings = map_info["standard_star_ratings"]
            self.taiko_sr_ratings = map_info["taiko_star_ratings"]
            self.ctb_sr_ratings = map_info["ctb_star_ratings"]
            self.mania_sr_ratings = map_info["mania_star_ratings"]
        self.drain_time = map_info["drain_time"]
        self.preview_start = map_info["preview_start_time"]
        self.timing_points = map_info["timing_points"]
        self.thread_id = map_info["thread_id"]
        self.best_grade_standard = map_info["best_grade_standard"]
        self.best_grade_taiko = map_info["best_grade_taiko"]
        self.best_grade_ctb = map_info["best_grade_ctb"]
        self.best_grade_mania = map_info["best_grade_mania"]
        self.local_offset = map_info["local_offset"]
        self.stack_leniency = map_info["stack_leniency"]
        self.online_offset = map_info["online_offset"]
        self.title_font = map_info["title_font"]
        self.unplayed = map_info["unplayed"]
        self.last_played = map_info["last_played"]
        self.is_osz2 = map_info["is_osz2"]
        self.folder_name = map_info["folder_name"]
        self.last_updated = map_info["last_updated"]
        self.ignore_sound = map_info["ignore_sound"]
        self.ignore_skin = map_info["ignore_skin"]
        self.disable_storyboard = map_info["disable_storyboard"]
        self.disable_video = map_info["disable_video"]
        self.visual_override = map_info["visual_override"]
        if map_info["mode"] == 3:
            self.scroll_speed = map_info["scroll_speed"]

        super().__init__(map_info, client)

    def get_online_map(self):
        return self.client.fetch_map(self.beatmap_id)


class Map(BaseMap):
    def __init__(self, map_info, client):
        super().__init__(map_info, client)
        self.bpm = float(map_info['bpm'])
        self.genre = Genre(int(map_info['genre_id']))
        self.language = Language(int(map_info['language_id']))
        self.creator_id = int(map_info['creator_id'])
        self.favourites = int(map_info['favourite_count'])
        self.rating = float(map_info['rating'])
        self.playcount = int(map_info['playcount'])
        self.passcount = int(map_info['passcount'])
        self.date_submitted = datetime.strptime(map_info['submit_date'], "%Y-%m-%d %H:%M:%S")
        self.date_approved = datetime.strptime(map_info['approved_date'], "%Y-%m-%d %H:%M:%S") if map_info[
            'approved_date'] else None
        self.last_update = datetime.strptime(map_info['last_update'], "%Y-%m-%d %H:%M:%S")
        self.star_rating = float(map_info['difficultyrating'])
        self.aim_difficulty = float(map_info['diff_aim'])
        self.speed_difficulty = float(map_info['diff_speed'])
        self.no_break_length = float(map_info['hit_length'])
        self.max_combo = int(map_info['max_combo'])
        self.storyboard = map_info['storyboard'] == "1"
        self.video = map_info['video'] == "1"
        self.download_unavailable = map_info['download_unavailable'] == "1"
        self.audio_unavailable = map_info['audio_unavailable'] == "1"
        self.download_url = f"https://osu.ppy.sh/osu/{self.beatmap_id}"
        self.cover_image_url = f"https://assets.ppy.sh/beatmaps/{self.mapset_id}/covers/cover.jpg"
        self.thumbnail = f"https://b.ppy.sh/thumb/{self.mapset_id}l.jpg"

    def fetch_creator(self):
        """Makes an api call to get more information about the map creator

        Returns:
            User: user who mapped the map
        """
        return self.client.get_user(self.creator_id)

    def fetch_mapset(self):
        """Makes an api call to get the rest of the maps in the beatmap set

        Returns:
            [Map]: list of maps in the mapset
        """
        return self.client.get_maps(set_id=self.mapset_id)

    def fetch_scores(self, username: str = None, user_id: int = None,
                     mode: Mode = Mode.STANDARD, mods: Mods = None, limit: int = 50):
        """Makes an api call to get scores for the map.

        Args:
            username: username of player to filter by
            user_id: id of player to filter by
            mode: mode to get the scores for
            mods: enum representing the mods to filter by
            limit: number of scores to return (max 100)

        Returns:
            [Score]: scores matching the search requirements on this map
        """
        return self.client.get_scores(self.beatmap_id, username=username, user_id=user_id,
                                      mode=mode, mods=mods, limit=limit)

    def get_pp(self, score=None, mods=Mods.NM):
        """
        Return the maximum PP value for the map with the supplied mods, or the PP for a given score on the map
        Args:
            score: Score to calculate PP for
            mods: enum representing mods to filter by

        Returns:
            float: Maximum PP for the map.
        """
        file = urllib.request.urlopen(self.download_url)
        map_input = []
        for line in file:
            map_input.append(line.decode().strip())
        p = pyttanko.parser()
        beatmap = p.map(map_input)
        if score:
            stars = pyttanko.diff_calc().calc(beatmap, mods=score.mods.value)
            pp_out, _, _, _, _ = pyttanko.ppv2(stars.aim, stars.speed, bmap=beatmap, mods=score.mods.value,
                                               max_combo=score.max_combo, n300=score.num_300, n100=score.num_100,
                                               n50=score.num_50, nmiss=score.misses)

        else:
            stars = pyttanko.diff_calc().calc(beatmap, mods=mods.value)
            pp_out, _, _, _, _ = pyttanko.ppv2(stars.aim, stars.speed, bmap=beatmap, mods=mods.value,
                                               max_combo=self.max_combo, n300=self.total_objects, n100=0,
                                               n50=0, nmiss=0)
        return pp_out



from .utils import *
from .map import LocalMap
from .score import LocalScore


class OsuDB:
    def __init__(self, client, path):
        self.client = client
        with open(path, "rb") as db:
            self.version = read_int(db)
            self.folder_count = read_int(db)
            self.account_unlocked = read_bool(db)
            self.date_unlocked = read_long(db)
            self.player_name = read_string(db)
            self.number_of_beatmaps = read_int(db)
            self.maps = {}
            self.maps_by_hash = {}
            for _ in range(self.number_of_beatmaps):
                map_info = {}
                if self.version < 20191106:
                    map_info["file_size"] = read_int(db)
                else:
                    map_info["file_size"] = -1
                map_info["osu_version"] = self.version
                map_info["artist"] = read_string(db)
                map_info["artist_unicode"] = read_string(db)
                map_info["title"] = read_string(db)
                map_info["title_unicode"] = read_string(db)
                map_info["creator"] = read_string(db)
                map_info["version"] = read_string(db)  # difficulty name
                map_info["audio_filename"] = read_string(db)
                map_info["file_md5"] = read_string(db)
                map_info["filename"] = read_string(db)
                ranked_status = read_byte(db)  # why is this different to the online api
                map_info["approved"] = ranked_status - 3 if ranked_status >= 4 else -2
                map_info["count_normal"] = read_short(db)
                map_info["count_slider"] = read_short(db)
                map_info["count_spinner"] = read_short(db)
                map_info["last_mod_time"] = read_long(db)
                if self.version < 20140609:
                    map_info["diff_approach"] = int(read_byte(db))
                    map_info["diff_size"] = int(read_byte(db))
                    map_info["diff_drain"] = int(read_byte(db))
                    map_info["diff_overall"] = int(read_byte(db))
                else:
                    map_info["diff_approach"] = read_single(db)
                    map_info["diff_size"] = read_single(db)
                    map_info["diff_drain"] = read_single(db)
                    map_info["diff_overall"] = read_single(db)
                map_info["slider_velocity"] = read_double(db)
                if self.version >= 20140609:
                    no_pairs = read_int(db)
                    map_info["standard_star_ratings"] = [read_int_double(db) for _ in range(no_pairs)]
                    no_pairs = read_int(db)
                    map_info["taiko_star_ratings"] = [read_int_double(db) for _ in range(no_pairs)]
                    no_pairs = read_int(db)
                    map_info["ctb_star_ratings"] = [read_int_double(db) for _ in range(no_pairs)]
                    no_pairs = read_int(db)
                    map_info["mania_star_ratings"] = [read_int_double(db) for _ in range(no_pairs)]
                map_info["drain_time"] = read_int(db)
                map_info["total_length"] = read_int(db)
                map_info["preview_start_time"] = read_int(db)
                no_timing = read_int(db)
                map_info["timing_points"] = [read_timing(db) for _ in range(no_timing)]
                map_info["beatmap_id"] = read_int(db)
                map_info["beatmapset_id"] = read_int(db)
                map_info["thread_id"] = read_int(db)
                map_info["best_grade_standard"] = read_byte(db)
                map_info["best_grade_taiko"] = read_byte(db)
                map_info["best_grade_ctb"] = read_byte(db)
                map_info["best_grade_mania"] = read_byte(db)
                map_info["local_offset"] = read_short(db)
                map_info["stack_leniency"] = read_single(db)
                map_info["mode"] = read_byte(db)
                map_info["source"] = read_string(db)
                map_info["tags"] = read_string(db)
                map_info["online_offset"] = read_short(db)
                map_info["title_font"] = read_string(db)
                map_info["unplayed"] = read_bool(db)
                map_info["last_played"] = read_long(db)
                map_info["is_osz2"] = read_bool(db)
                map_info["folder_name"] = read_string(db)
                map_info["last_updated"] = read_long(db)
                map_info["ignore_sound"] = read_bool(db)
                map_info["ignore_skin"] = read_bool(db)
                map_info["disable_storyboard"] = read_bool(db)
                map_info["disable_video"] = read_bool(db)
                map_info["visual_override"] = read_bool(db)
                if self.version < 20140609:
                    read_short(db)  # wiki says this just isn't actually a used value
                map_info["last_mod_time_2"] = read_int(db)
                map_info["scroll_speed"] = read_byte(db)
                mp = LocalMap(map_info, client)
                self.maps[mp.beatmap_id] = mp
                self.maps_by_hash[mp.md5_hash] = mp

    def search_maps(self, name: str):
        results = []
        name = name.lower()
        for local_map in self.maps.values():
            if name in local_map.song_title.lower():
                results.append(local_map)
        return results

    def get_map_from_hash(self, md5_hash: str):
        for local_map in self.maps.values():
            if local_map.md5_hash == md5_hash:
                return local_map
        return None


class Collections:
    def __init__(self, path, client):
        self.client = client
        with open(path, "rb") as db:
            self.version = read_int(db)
            self.collections = {}
            no_collections = read_int(db)
            for _ in range(no_collections):
                collection = []
                name = read_string(db)
                no_maps = read_int(db)
                for _ in range(no_maps):
                    collection.append(read_string(db))
                self.collections[name] = collection


class ScoresDB:
    def __init__(self, path, client):
        self.client = client
        with open(path, "rb") as db:
            self.version = read_int(db)
            no_maps = read_int(db)
            self.maps = {}
            for _ in range(no_maps):
                md5 = read_string(db)
                scores = []
                no_scores = read_int(db)
                for _ in range(no_scores):
                    score_info = {"mode": read_byte(db), "version": read_int(db)}
                    map_md5 = read_string(db)
                    score_info["username"] = read_string(db)
                    score_info["replay_hash"] = read_string(db)
                    score_info["count300"] = read_short(db)
                    score_info["count100"] = read_short(db)
                    score_info["count50"] = read_short(db)
                    score_info["countgeki"] = read_short(db)
                    score_info["countkatu"] = read_short(db)
                    score_info["countmiss"] = read_short(db)
                    score_info["score"] = read_int(db)
                    score_info["maxcombo"] = read_short(db)
                    score_info["perfect"] = read_bool(db)
                    score_info["mods"] = read_int(db)
                    read_string(db)
                    score_info["timestamp"] = read_long(db)
                    read_int(db)
                    score_info["online_id"] = read_long(db)
                    if Mods.Target & Mods(score_info["mods"]):
                        score_info["target_practice_acc"] = read_double(db)

                    scores.append(LocalScore(score_info, client, score_info["replay_hash"]))

                self.maps[md5] = scores




import requests
from .user import User
from .map import Map
from .score import Score, RecentScore
from .utils import *
from .match import Match
from .db import OsuDB, Collections, ScoresDB


class OsuClient:
    """Client object for interacting with the osu! api.

    Attributes:
        api_key (str): the api key for your osu! application

    """
    def __init__(self, key: str):
        """Initialise the client with an api key.

        Args:
            key: The api key for your osu! application.
        """
        self.api_key = key
        self.osu_db = None
        self.collections_db = None
        self.scores_db = None

    @staticmethod
    def _id_or_name(params: dict, username: str, user_id: int):
        """Configured a given parameter dict depending on if it has received a valid username or user_id.

        Args:
            params: dict of parameters to be sent
            username: username or None if given a user_id
            user_id: user id or None if given a username

        Returns:
            None
        """
        if username:
            params['u'] = username
            params['type'] = "string"
        elif user_id:
            params['u'] = user_id
            params['type'] = "id"

    def _request_api(self, url: str, params: dict):
        """Private method to generalise api requests.

        Args:
            url: the api endpoint to request to (e.g. "get_user")
            params: dict of prepared parameters for the endpoint you are attempting to request

        Returns:
            dict: results of the api request (or None if unsuccessful)
        """
        params['k'] = self.api_key
        r = requests.get(f"https://osu.ppy.sh/api/{url}", params=params)
        return r.json() if r.json() else None

    def fetch_user(self, user_id: int = None, username: str = None, mode: Mode = Mode.STANDARD):
        """Make an api request to get information about a given user.

        Args:
            user_id: id of the user
            username: name of the user
            mode: enum representing the osu! gamemode you want the information for

        Returns:
            osupy.User: User object representing the requested user
        """
        if not (username or user_id):
            return
        params = {"m": mode.value}
        self._id_or_name(params, username, user_id)
        user_json = self._request_api("get_user", params)
        if user_json:
            return User(user_json[0], self)

    def fetch_scores(self, map_id: int, username: str = None, user_id: int = None,
                     mode: Mode = Mode.STANDARD, mods: Mods = None, limit: int = 50):
        """Make an api request to retrieve information about scores on a given beatmap.

        Args:
            map_id: the id of the beatmap to retrieve scores for
            username: username of a player to filter by
            user_id: id of a player to filter by
            mode: enum representing the osu! gamemode you want information for
            mods: enum representing the mod combination to filter by
            limit: number of scores to retrieve (max 100)

        Returns:
            [Score]: List of Score objects containing information about a given score
        """
        params = {"b": map_id, "m": mode.value}
        self._id_or_name(params, username, user_id)
        if mods:
            params['mods'] = mods.value
        if 1 <= limit <= 100:
            params['limit'] = limit

        scores_json = self._request_api("get_scores", params)
        if scores_json:
            return [Score(score_info, self, params['b']) for score_info in scores_json]

    def fetch_map(self, map_id: int):
        """Wrapper for the api call of selecting a map from an id and taking the first one, as it would only return one map anyway.

        Args:
            map_id: the map to get information for

        Returns:
            Map: Map object containing information about the map requested
        """
        return self.fetch_maps(map_id=map_id)[0]

    # todo: add error handling if invalid mods given
    def fetch_maps(self, set_id: int = None, map_id: int = None, username: str = None, user_id: int = None,
                   map_hash: str = None, mode: Mode = Mode.STANDARD, converts: bool = False,
                   limit: int = 500, mods: Mods = Mods.NM, since: datetime = None):
        """Api call to search osu!'s beatmap pool.

        Args:
            set_id: the id of a beatmapset to filter by
            map_id: the id of a specific map to return
            username: filter by mapper's username
            user_id: filter by mapper's id
            map_hash: the md5 hash of a beatmap
            mode: enum representing the osu! gamemode to filter by
            converts: whether to include converted maps for a gamemode
            limit: number of maps to return (max and default 500)
            mods: enum representing the mod combination to filter by
            since: limit beatmaps found to between this date and the present

        Returns:
            [Map]: List of Map objects containing information about the maps requested
        """
        params = {"m": mode.value, "mods": mods.value}
        self._id_or_name(params, username, user_id)
        if set_id:
            params['s'] = set_id
        if map_id:
            params['b'] = map_id
        if map_hash:
            params['h'] = map_hash
        if converts:
            params['a'] = 1
        if 0 <= limit <= 500:
            params['limit'] = limit
        if since:
            params['since'] = since.strftime("%Y-%m-%d %H:%M:%S")

        maps_json = self._request_api("get_beatmaps", params)
        if maps_json:
            return [Map(map_info, self) for map_info in maps_json]

    def fetch_user_best(self, username: str = None, user_id: int = None, mode: Mode = Mode.STANDARD, limit: int = 10):
        """Api request to find a users best scores.

        Args:
            username: username of user
            user_id: id of a user
            mode: enum representing mode to get the top plays from
            limit: number of best scores to retrieve (max 100)

        Returns:
            [Score]: List of Score objects representing the user's best plays
        """
        if not (username or user_id):
            return
        params = {"m": mode.value}
        if 1 <= limit <= 100:
            params['limit'] = limit
        self._id_or_name(params, username, user_id)

        best_json = self._request_api("get_user_best", params)
        if best_json:
            return [Score(score_info, self, score_info['beatmap_id']) for score_info in best_json]

    def fetch_user_recent(self, username: str = None, user_id: int = None, mode: Mode = Mode.STANDARD, limit: int = 10):
        """Api request to retrieve a user's recent scores (within the last 24 hours).

        Args:
            username: username of user
            user_id: id of user
            mode: enum representing osu! gamemode to retrieve scores for
            limit: number of scores to retrieve (max 100)

        Returns:
            [Score]: List of Score objects representing a user's recent plays
        """
        if not (username or user_id):
            return
        params = {"m": mode.value}
        self._id_or_name(params, username, user_id)
        if 1 <= limit <= 100:
            params['limit'] = limit
        recent_json = self._request_api("get_user_recent", params)
        if recent_json:
            return [RecentScore(score_info, self, score_info['beatmap_id']) for score_info in recent_json]

    """
    Requesting via beatmap and user was broken when I was testing but I'll leave the function here in case it's just
    me being dumb
    def fetch_replay(self,  score_id: int = None, username: str = None, user_id: int = None,
                     mode: Mode = Mode.STANDARD, map_id: int = None, mods: Mods = Mods.NM):
        params = {}
        if score_id:
            params['s'] = score_id
        else:
            if not (username or user_id) or not map_id:
                return
            _id_or_name(params, username, user_id)
            params['b'] = map_id
            params['m'] = mode.value
            params['mods'] = mods.value
        return self._request_api("get_replay", params)
    
    """

    def fetch_replay(self, score_id: int):
        """Api request to fetch a replay for a given score

        Args:
            score_id: score to retrieve replay for

        Returns:
            str: base64 representation of LZMA stream, not .osr file. See https://osu.ppy.sh/help/wiki/osu!_File_Formats/Osr_(file_format) for details
        """
        replay_json = self._request_api("get_replay", {"s": score_id})
        return replay_json["content"] if "content" in replay_json else None

    def fetch_match(self, match_id: int):
        """Api request to fetch details about a given multiplayer lobby, past or ongoing

        Args:
            match_id: id of the match

        Returns:
            Match: Match object representing all the information available about the match requested.
        """
        match_json = self._request_api("get_match", {"mp": match_id})
        return Match(match_json, self)

    def get_local_map(self, md5_hash):
        if not self.osu_db:
            return None
        if md5_hash in self.osu_db.maps_by_hash.keys():
            return self.osu_db.maps_by_hash[md5_hash]
        return None

    def set_osu_db(self, path):
        self.osu_db = OsuDB(self, path)
        return self.osu_db

    def set_collections_db(self, path):
        self.collections_db = Collections(path, self)
        return self.collections_db

    def set_scores_db(self, path):
        self.scores_db = ScoresDB(path, self)
        return self.scores_db



from datetime import datetime, timedelta
from .utils import Mode, Playtime


class User:
    """Object representing JSON response from the osu! api plus some helper methods.

    Attributes:
        client: OsuClient that created the instance
        mode: enum representing the osu! gamemode this information is for
        id: id of the player
        username: player's username
        join_date: timestamp of when the player created their account
        num_300: number of 300s player has achieved
        num_100: number of 100s player has achieved
        num_50: number of 50s player has achieved
        play_count: the number of "plays" according to however osu! calculates it
        ranked_score: combination of highest submitted scores on ranked maps
        total_score: combination of highest submitted scores across all online maps
        rank: the position of the player on the global leaderboard based on their pp
        level: the player's level
        pp: the player's pp total (points received from ranked maps)
        accuracy: the player's average accuracy on a map as a percentage
        ss_count: number of SS's
        ssh_count: number of SS's with hidden
        s_count: number of S ranks
        sh_count: number of S ranks with hidden
        a_count: number of A ranks
        country: the country the player is from
        playtime: Playtime object representing how long they have played the game
        recents: dict containing a list of cached recent scores for each mode
        avatar_url: the url for their profile picture
    
    """
    def __init__(self, user_info, osu_client, mode: Mode = Mode.STANDARD):
        self.client = osu_client
        self.mode = mode
        self.id = int(user_info['user_id'])
        self.username = user_info['username']
        self.join_date = datetime.strptime(user_info['join_date'], "%Y-%m-%d %H:%M:%S")
        self.num_300 = int(user_info['count300'])
        self.num_100 = int(user_info['count100'])
        self.num_50 = int(user_info['count50'])
        self.play_count = int(user_info['playcount'])
        self.ranked_score = int(user_info['ranked_score'])
        self.total_score = int(user_info['total_score'])
        self.rank = int(user_info['pp_rank'])
        self.level = float(user_info['level'])
        self.pp = float(user_info['pp_raw'])
        self.accuracy = float(user_info['accuracy'])
        self.ss_count = int(user_info['count_rank_ss'])
        self.ssh_count = int(user_info['count_rank_ssh'])
        self.s_count = int(user_info['count_rank_s'])
        self.sh_count = int(user_info['count_rank_sh'])
        self.a_count = int(user_info['count_rank_a'])
        self.country = user_info['country']
        self.country_rank = int(user_info['pp_country_rank'])
        seconds = int(user_info['total_seconds_played'])
        self.playtime = Playtime(seconds)
        self.recents = {
            Mode.STANDARD: [],
            Mode.TAIKO: [],
            Mode.CTB: [],
            Mode.MANIA: []
        }
        self.avatar_url = f"http://s.ppy.sh/a/{self.id}"

    def fetch_best(self, mode: Mode = Mode.STANDARD, limit: int = 10):
        """Helper for collecting a user's best scores.

        Args:
            mode: enum representing osu! gamemode to request scores for
            limit: number of scores to retrieve (max 100)

        Returns:
            [Score]: List of the player's best scores
        """
        return self.client.fetch_user_best(user_id=self.id, mode=mode, limit=limit)

    def fetch_recent(self, mode: Mode = Mode.STANDARD, limit: int = 10):
        """Helper for collecting a user's recent scores.

        Args:
            mode: enum representing osu! gamemode to request scores for
            limit: number of scores to retrieve (max 100)

        Returns:
            [RecentScore]: List of the player's recent scores
        """
        self.recents[mode] = self.client.fetch_user_recent(user_id=self.id, mode=mode, limit=limit)
        return self.recents[mode]

    def fetch_new_recent(self, mode: Mode = Mode.STANDARD):
        """Retrieve any scores made since the last time this function was called.

        Args:
            mode: enum representing osu! gamemode to request scores for

        Returns:
            [RecentScore]: List of the player's scores since the most recent
        """
        now = datetime.now()
        self.recents[mode] = [score for score in self.recents[mode] if (now - timedelta(hours=24)) <= score.timestamp <= now]
        new_scores = self.client.fetch_user_recent(user_id=self.id, mode=mode, limit=100)
        if new_scores:
            dates = [score.timestamp for score in self.recents[mode]]
            new = [score for score in new_scores if score.timestamp not in dates]
            self.recents[mode] += new
            return new

    def __repr__(self):
        return self.username

